/*============================================================================
                            L Q M E N U (version 1.1)

                        S.M. Willett - September 9, 1990

   LQMENU allows the user to control many of the features of the Epson
   family of LQ printers.  The program consists of three menus:
     Main Menu  - This menu allows the printer to be configured with many
                  different font pitches, underlining, bold, letter/draft
                  quality, etc...
     Font Menu  - This menu displays up to eighty downloadable fonts in the
                  current directory or the directory specified in the setup
                  menu.  When the highlighted bar is placed over the
                  desired font and Alt-d is pressed, the font is downloaded
                  to the printer.  The font file directory and extension
                  are those specified in the setup menu.
     Setup Menu - This menu allows the user to specify the directory for
                  the downloadable fonts, the font filenames extension,
                  carriage size, printer port, and initialization.  It will
                  create a data file for the program to read when invoked.
============================================================================*/


/*============================================================================
INCLUDES
----------------------------------------------------------------------------*/
#include <bios.h>
#include <conio.h>
#include <ctype.h>
#include <dir.h>
#include <dos.h>
#include <io.h>
#include <process.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>


/*============================================================================
DEFINES
----------------------------------------------------------------------------*/
#define BACKSLASH  92
#define ESC  27
#define HIB  0xFF
#define HORZ  205
#define LCOL  35
#define LL  200
#define LR  188
#define NO_MAIN_ITEMS  18
#define NO_SETUP_ITEMS  5
#define OFF  0
#define ON  1
#define RCOL  75
#define T2L  185
#define T2R  204
#define TDOWN  203
#define TUP  202
#define UL  201
#define UR  187
#define VERT  186


/*============================================================================
DATA DECLARATIONS
----------------------------------------------------------------------------*/
typedef unsigned char BYTE ;
BYTE init_code[3] = { ESC,'@',NULL } ;
char choice, initialize ;
char filename[65], font_dir[55], font_ext[6], pgm_dir[63] ;
int black, blue, carriage, chg_dir, cyan, gray, green, hi_red ;
int max_margin, no_fonts, old_attr, prn_port, red, white, yellow ;
struct { char fontname[9] ;
         int len          ;
         int x            ;
         int y            ; } font_tbl[80] ;
struct { int x         ;
         int y         ;
         int ndx       ;
         int max_bytes ;
         BYTE code[4]  ; } main_item[NO_MAIN_ITEMS] = {
  { LCOL, 8, 3,2,{ESC,'P',HIB,HIB} },    /* CPI - Pica Width */
  { LCOL, 9, 5,1,{ESC,'W',  0,HIB} },    /* Double Width - Cancel */
  { LCOL,10, 8,1,{18 ,HIB,HIB,HIB} },    /* Condensed - Cancel */
  { LCOL,13, 9,1,{ESC,'5',HIB,HIB} },    /* Italics - Cancel */
  { LCOL,14, 0,1,{ESC,'p',  0,HIB} },    /* Proportional Character - Cancel */
  { LCOL,15,11,3,{ESC,'T',  0,HIB} },    /* Superscript/Subscript - Cancel */
  { LCOL,18,14,3,{ESC,'l',  0,HIB} },    /* Left Margin - Column 0 */
  { LCOL,19,17,3,{ESC,'Q', 80,HIB} },    /* Right Margin - Column 80 */
  { RCOL, 8,20,2,{ESC,'x',  0,HIB} },    /* Draft/LQ - Draft Quality */
  { RCOL, 9,23,1,{ESC,'H',HIB,HIB} },    /* Double Strike - Cancel */
  { RCOL,10,25,1,{ESC,'F',HIB,HIB} },    /* Emphasized - Cancel */
  { RCOL,11,27,1,{ESC,'U',  0,HIB} },    /* Unidirectional - Cancel */
  { RCOL,12,30,1,{ESC,'s',  0,HIB} },    /* Half-Speed - Cancel */
  { RCOL,15,33,2,{ESC,'C',  0, 11} },    /* Page Length - 11" */
  { RCOL,16,37,3,{ESC,'C', 66,HIB} },    /* Lines per Page - 66 Lines */
  { RCOL,17,40,1,{ESC,'A', 10,HIB} },    /* 1/8" Line Spacing - 10/60" */
  { RCOL,18,40,2,{ESC,'A', 10,HIB} },    /* Line Spacing n/60" - 10/60" */
  { RCOL,19,43,3,{ESC,'N',  0,HIB} } } ; /* Skip Lines B/Perf - 0 lines */
struct { int x ;
         int y ;
         int max_bytes ; } setup_item[NO_SETUP_ITEMS] =   /* col, row, & len */
  { {27,7,52},{27,10,3},{31,13,1},{38,16,2},{70,19,1} } ; /* of setup items  */
struct record { BYTE prop_char[3] ;
                BYTE pitch[2] ;
                BYTE dw[3] ;
                BYTE cond[1] ;
                BYTE italic[2] ;
                BYTE sup_sub[3] ;
                BYTE left_mar[3] ;
                BYTE right_mar[3] ;
                BYTE draft_lq[3] ;
                BYTE ds[2] ;
                BYTE emph[2] ;
                BYTE uni[3] ;
                BYTE half_speed[3] ;
                BYTE length[4] ;
                BYTE lines[3] ;
                BYTE line_spacing[3] ;
                BYTE skip_perf[3] ;    } ;
struct record workrec ;
union REGS Regs ; /* Used by functions for DOS calls */
unsigned int scan_codes[9] = /* hi bit corresponds to allowable scan code */
  { 0x0000, 0x8000, 0x4003, 0x0004, 0x8180, 0x0001, 0x0000, 0x0000, 0x0000 } ;


/*============================================================================
FUNCTION DECLARATIONS
----------------------------------------------------------------------------*/
void assign_default (int item_num) ;
int  config (char mode) ;
int  cursor (int on_off,int cur_size) ;
void display_font_menu (void) ;
int  display_init (void) ;
void display_main_menu (void) ;
void display_setup_menu (void) ;
int  file2prn (char *fname) ;
void font_menu (void) ;
void horz_line (void) ;
int  input_error (void) ;
void load_font_tbl (void) ;
void main_menu (void) ;
void menu_frame (char *title) ;
void message (char *msg) ;
void no_yes (BYTE *workrec_ptr,char no_value) ;
void other_color (char *str,int color) ;
void outside_edges (int line) ;
void position (int old,int new) ;
void post_limits (char *first,char *second) ;
void post_yn (void) ;
int  prn_status (int port) ;
extern REVID(unsigned int x,unsigned int y,unsigned int no_chars) ;
void setup_menu (void) ;
void tee_line (int line) ;
void update_main (int item_num) ;
void update_setup (int item_num) ;
unsigned int user_input (char *str,int max_bytes) ;


/*============================================================================
FUNCTION NAME:  main
PURPOSE:  Controls which menu is active, and if ESCAPE is pressed, exits the
          program.  Also exits if monitor is 40 column.
RETURNS:  0 in all cases
ARGUMENTS:  int argc    = number of arguments on command line
            **char argv = array of argument strings
GLOBAL VARIABLES:  chg_dir, choice, old_attr, pgm_dir
FUNCTIONS USED:  clrscr, config, cursor, display_init, font_menu, main_menu,
                 setup_menu, strcpy, strlen, textattr
----------------------------------------------------------------------------*/
void main ( int argc, char **argv )
{
int first_time, ndx, old_cur_size ;

  if ( display_init() ) /* 80 column monitor */
  {
    strcpy(pgm_dir,argv[0]) ; /* find program path */
    for (ndx=strlen(pgm_dir)-1 ; ndx>0 ; ndx--)
    {
      if ( pgm_dir[ndx] == BACKSLASH )
        break ;
      else
          pgm_dir[ndx] = NULL ;
    }
    first_time = config('L') ; /* load configuration data */
    choice = ( first_time==1 ? 's':'m' ) ; /* display setup menu if no data */
                                           /* file, else display main menu  */
    chg_dir = 1 ; /* ensure font names are read initially */
    old_cur_size = cursor(OFF,NULL) ;
    while ( choice != ESC )
    {
      switch ( choice )
      {
        case 'f' : font_menu() ; /* Font Menu */
                   break ;
        case 'm' : main_menu() ; /* Main Menu */
                   break ;
        case 's' : setup_menu() ; /* Setup Menu */
                   break ;
        default  : break ;
      }
    }
    textattr(old_attr) ; /* Restore old screen attributes */
    clrscr() ;
    cursor(ON,old_cur_size) ;
  }
}


/*============================================================================
FUNCTION NAME:  assign_default
PURPOSE:  Stores the default code in workrec
RETURNS:  void
ARGUMENTS:  int item_num = item number to be assigned default code
GLOBAL VARIABLES:  main_item, workrec
FUNCTIONS USED:  None
----------------------------------------------------------------------------*/
void assign_default ( int item_num )
{
int ndx ;
BYTE *workrec_ptr ;

  workrec_ptr = &workrec.prop_char[0] ;
  for (ndx=0 ; ndx<4 ; ndx++)
  {
    if ( main_item[item_num].code[ndx] == HIB )
      break ;
    else
      *(workrec_ptr + main_item[item_num].ndx + ndx) =
                    main_item[item_num].code[ndx] ;
  }
}


/*============================================================================
FUNCTION NAME:  config
PURPOSE:  Loads or saves configuration data in file lqmenu.dat.  If file
          doesn't exist, sets configuration variables to default.
RETURNS:  int = 0, successful
                1, error opening configuration file or file doesn't exist
ARGUMENTS:  char mode = L, load data file
                        S, save data file
GLOBAL VARIABLES:  carriage, filename, font_dir, initialize, pgm_dir, prn_port,
                   workrec
FUNCTIONS USED:  fclose, fgetc, fopen, fprintf, fputc, fscanf, getw, putw,
                 strcpy, strcat
----------------------------------------------------------------------------*/
int config ( char mode )
{
FILE *datfile ; /* file pointer to file lqmenu.dat */
BYTE *workrec_ptr ; /* pointer to workrec */
int ndx ; /* a counter */
int return_value = 0 ; /* value returned to calling routine */

  return_value = 0 ;
  workrec_ptr = &workrec.prop_char[0] ;
  strcpy(filename,pgm_dir) ;
  strcat(filename,"LQMENU.DAT") ;
  if ( mode == 'L' )
  {
    datfile = fopen(filename,"rb") ;
    if ( datfile != NULL ) /* data file exists - retrieve data */
    {
      fgets(font_dir,sizeof(font_dir)-1,datfile) ;
      font_dir[strlen(font_dir)-1] = NULL ; /* remove newline character */
      fgets(font_ext,sizeof(font_ext)-1,datfile) ;
      font_ext[strlen(font_ext)-1] = NULL ; /* remove newline character */
      prn_port = getw(datfile) ;
      carriage = getw(datfile) ;
      if ( carriage == 15 ) /* 15" carriage */
        main_item[7].code[2] = 136 ;
      initialize = fgetc(datfile) ;
      fread(workrec_ptr,sizeof(workrec),1,datfile) ;
    }
    else /* assign default values */
    {
      strcpy(font_dir,pgm_dir) ;
      strcpy(font_ext,"FNT") ;
      prn_port = 0 ;
      carriage = 10 ;
      initialize = 'N' ;
      for (ndx=0 ; ndx<NO_MAIN_ITEMS ; ndx++ )
        assign_default(ndx) ;
      return_value++;
    }
  }
  else
  {
    if ( mode == 'S' ) /* save data file */
    {
      datfile = fopen(filename,"wb") ;
      if ( datfile != NULL ) /* file opened okay - save data */
      {
        fprintf(datfile,"%s\n%s\n",font_dir,font_ext) ;
        putw(prn_port,datfile) ;
        putw(carriage,datfile) ;
        fputc(initialize,datfile) ;
        fwrite(workrec_ptr,sizeof(workrec),1,datfile) ;
      }
      else /* error opening data file */
        return_value++ ;
    }
  }
  fclose(datfile) ;
  return(return_value) ;
}


/*============================================================================
FUNCTION NAME:  cursor
PURPOSE:  Turns cursor on or off
RETURNS:  int = cursor size when turning cursor off
ARGUMENTS:  int on_off =   0, turn cursor off
                           1, turn cursor on using cur_size
            int cur_size = cursor size when turning cursor on
GLOBAL VARIABLES:  Regs
FUNCTIONS USED:  int86
----------------------------------------------------------------------------*/
int cursor ( int on_off, int cur_size )
{
  if ( on_off == ON ) /* turn cursor on */
  {
    Regs.x.cx = cur_size ;
    Regs.h.ah = 0x01 ;
    int86(0x10,&Regs,&Regs) ;
  }
  else /* turn cursor off */
  {
    Regs.h.ah = 0x03 ; /* get current cursor size */
    int86(0x10,&Regs,&Regs) ;
    cur_size = Regs.x.cx ;
    Regs.h.ch = 32 ; /* blank cursor */
    Regs.h.ah = 0x01 ;
    int86(0x10,&Regs,&Regs) ;
  }
  return(cur_size) ;
}


/*============================================================================
FUNCTION NAME:  display_font_menu
PURPOSE:  Displays the font menu
RETURNS:  void
ARGUMENTS:  void
GLOBAL VARIABLES:  black, gray, red, white
FUNCTIONS USED:  cputs, gotoxy, menu_frame, other_color, textbackground,
                 textcolor
----------------------------------------------------------------------------*/
void display_font_menu ( void )
{
  textbackground(gray) ; /* Gray background */
  textcolor(red) ; /* Red frame */
  menu_frame("L Q M E N U   D O W N L O A D A B L E   F O N T S") ;
  textcolor(black) ; /* Black text */
  /* ... Line 23 ... */
  gotoxy(8,23) ;
  other_color("ESC",white) ;
  cputs(" to exit       ") ;
  other_color("Alt-D",white) ;
  cputs("ownload font       ") ;
  other_color("Alt-M",white) ;
  cputs("ain       ") ;
  other_color("Alt-S",white) ;
  cputs("etup") ;
}


/*============================================================================
FUNCTION NAME:  display_init
PURPOSE:  Determines display mode and if 80 column saves the current screen
          attributes.  Sets colors if color monitor.
RETURNS:  int = 0, 40 column (program won't run)
                1, 80 column (program will run)
ARGUMENTS:  void
GLOBAL VARIABLES:  black, blue, cyan, gray, green, hi_red, old_attr, red,
                   text_info, white, yellow
FUNCTIONS USED:  cputs, gettextinfo, int86
----------------------------------------------------------------------------*/
int display_init ( void )
{
struct text_info Old_display_data ; /* Contains DOS screen attributes */
int return_value = 1 ;

  Regs.h.ah = 0x0F ; /* Get current video mode */
  int86(0x10,&Regs,&Regs) ;
  if ( Regs.h.al < 2 ) /* 40 column display */
  {
    cputs("\n\nNeed 80 column monitor for program\n") ;
    return_value-- ;
  }
  else
  {
    gettextinfo(&Old_display_data) ; /* Get current screen's attribute */
    old_attr = Old_display_data.attribute ; /* Save old attribute */
    switch ( Regs.h.al ) /* Display mode */
    {
      case  2 :               /* Black and white monitor */
      case  7 :
      case 15 : black  =  0 ;
                blue   =  0 ;
                cyan   = 15 ;
                red    =  7 ;
                gray   =  7 ;
                green  =  7 ;
                yellow =  7 ;
                white  = 15 ;
                hi_red =  7 ;
                break ;
      default : black  =  0 ; /* Color monitor */
                blue   =  1 ;
                cyan   = 11 ;
                red    =  4 ;
                gray   =  7 ;
                green  = 10 ;
                yellow = 14 ;
                white  = 15 ;
                hi_red = 12 ;
                break ;
    }
  }
  return(return_value) ;
}


/*============================================================================
FUNCTION NAME:  display_main_menu
PURPOSE:  Displays the main menu
RETURNS:  void
ARGUMENTS:  void
GLOBAL VARIABLES:  blue, gray, green, max_margin, white, yellow
FUNCTIONS USED:  cputs, gotoxy, menu_frame, other_color, post_limits, post_yn,
                 putch, sprintf, textbackground, textcolor
----------------------------------------------------------------------------*/
void display_main_menu ( void )
{
char outstr[4] ;

  sprintf(outstr,"%3d",max_margin) ;
  textbackground(blue) ; /* Blue background */
  textcolor(yellow) ; /* Bright yellow text */
  menu_frame("L Q M E N U  Version 1.1") ;
  gotoxy(40,5) ; /* Display tee-down character */
  putch(TDOWN) ;
  /* ... Line 6 ... */
  gotoxy(40,6) ; /* Display vertical double line */
  putch(VERT) ;
  /* ... Line 7 ... */
  gotoxy(3,7) ;
  other_color("FONT PITCH:",green) ; /* High intensity green */
  gotoxy(40,7) ;
  putch(VERT) ;
  other_color(" PRINT MODE:",green) ;
  /* ... Line 8 ... */
  gotoxy(4,8) ;
  cputs("10,12,15 CPI (") ;
  other_color("10",white) ; /* High intensity white */
  putch('/') ;
  other_color("12",white) ;
  putch('/') ;
  other_color("15",white) ;
  cputs(") ......") ;
  gotoxy(40,8) ;
  putch(VERT) ;
  cputs("  Draft/LQ (") ;
  other_color("D",white) ;
  putch('/') ;
  other_color("LQ",white) ;
  cputs(") ...............") ;
  /* ... Line 9 ... */
  gotoxy(4,9) ;
  cputs("Double Width ") ;
  post_yn() ;
  cputs("...........") ;
  gotoxy(40,9) ;
  putch(VERT) ;
  cputs("  Double Strike ") ;
  post_yn() ;
  cputs("...........") ;
  /* ... Line 10 ... */
  gotoxy(4,10) ;
  cputs("Condensed <x1.7> ") ;
  post_yn() ;
  cputs(".......") ;
  gotoxy(40,10) ;
  putch(VERT) ;
  cputs("  Emphasized ") ;
  post_yn() ;
  cputs("..............") ;
  /* ... Line 11 ... */
  gotoxy(40,11) ;
  putch(VERT) ;
  cputs("  Unidirectional ") ;
  post_yn() ;
  cputs("..........") ;
  /* ... Line 12 ... */
  gotoxy(3,12) ;
  other_color("FONT STYLE:",green) ;
  gotoxy(40,12) ;
  putch(VERT) ;
  cputs("  Half-speed ") ;
  post_yn() ;
  cputs("..............") ;
  /* ... Line 13 ... */
  gotoxy(4,13) ;
  cputs("Italics ") ;
  post_yn() ;
  cputs("................") ;
  gotoxy(40,13) ;
  putch(VERT) ;
  /* ... Line 14 ... */
  gotoxy(4,14) ;
  cputs("Prop Char ") ;
  post_yn() ;
  cputs("..............") ;
  gotoxy(40,14) ;
  putch(VERT) ;
  other_color(" VERTICAL COMMANDS:",green) ;
  /* ... Line 15 ... */
  gotoxy(4,15) ;
  cputs("Super/Subscript (") ;
  other_color("SUP",white) ;
  putch('/') ;
  other_color("SUB",white) ;
  putch('/') ;
  other_color("N",white) ;
  cputs(") ..") ;
  gotoxy(40,15) ;
  putch(VERT) ;
  cputs("  Page Length ") ;
  putch(34) ; /* Quotation mark */
  putch(' ') ;
  post_limits("1","32") ;
  cputs("..........") ;
  /* ... Line 16 ... */
  gotoxy(40,16) ;
  putch(VERT) ;
  cputs("  Lines per Page ") ;
  post_limits("1","127") ;
  cputs("........") ;
  /* ... Line 17 ... */
  gotoxy(3,17) ;
  other_color("HORIZONTAL COMMANDS:",green) ;
  gotoxy(40,17) ;
  putch(VERT) ;
  cputs("  1/8") ;
  putch(34) ; /* Quotation mark */
  cputs(" Line Spacing ") ;
  post_yn() ;
  cputs(".......") ;
  /* ... Line 18 ... */
  gotoxy(4,18) ;
  cputs("Left Margin ") ;
  post_limits("0",outstr) ;
  cputs("..........") ;
  gotoxy(40,18) ;
  putch(VERT) ;
  cputs("  Line Spacing n/60") ;
  putch(34) ; /* Quotation mark */
  putch(' ') ;
  post_limits("1","85") ;
  cputs(".....") ;
  /* ... Line 19 ... */
  gotoxy(4,19) ;
  cputs("Right Margin ") ;
  post_limits("0",outstr) ;
  cputs(".........") ;
  gotoxy(40,19) ;
  putch(VERT) ;
  cputs("  Skip Lines B/Perf ") ;
  post_limits("0","255") ;
  cputs(".....") ;
  /* ... Line 20 ... */
  gotoxy(40,20) ;
  putch(VERT) ;
  /* ... Line 21 ... */
  gotoxy(40,21) ;
  putch(VERT) ;
  /* ... Line 22 ... */
  gotoxy(40,22) ;
  putch(TUP) ;
  /* ... Line 23 ... */
  textcolor(gray) ;
  gotoxy(4,23) ;
  other_color("ESC",white) ;
  cputs(" to exit   ") ;
  other_color("Alt-C",white) ;
  cputs("onfigure printer   ") ;
  other_color("Alt-D",white) ;
  cputs("efaults   ") ;
  other_color("Alt-F",white) ;
  cputs("onts   ") ;
  other_color("Alt-S",white) ;
  cputs("etup") ;
  textcolor(gray) ;
}


/*============================================================================
FUNCTION NAME:  display_setup_menu
PURPOSE:  Displays the setup menu
RETURNS:  void
ARGUMENTS:  void
GLOBAL VARIABLES:  black, cyan, gray, white,
FUNCTIONS USED:  cputs, gotoxy, menu_frame, other_color, post_yn, putch
                 textbackground, textcolor
----------------------------------------------------------------------------*/
void display_setup_menu ( void )
{
  textbackground(black) ; /* Black background */
  textcolor(cyan) ; /* Cyan frame */
  menu_frame("L Q M E N U   S E T U P") ;
  textcolor(gray) ; /* Cyan text */
  /* ... Line 7 ... */
  gotoxy(3,setup_item[0].y) ;
  cputs("Font Files Directory =>") ;
  /* ... Line 10 ... */
  gotoxy(3,setup_item[1].y) ;
  cputs("Font Files Extension =>") ;
  /* ... Line 13 ... */
  gotoxy(3,setup_item[2].y) ;
  cputs("Printer Port (LPT") ;
  other_color("1",white) ;
  cputs("/LPT") ;
  other_color("2",white) ;
  cputs(") =>") ;
  /* ... Line 16 ... */
  gotoxy(3,setup_item[3].y) ;
  cputs("Printer Carriage Size (") ;
  other_color("10",white) ;
  putch(34) ; /* " */
  cputs("/") ;
  other_color("15",white) ;
  putch(34) ;
  cputs(") =>") ;
  /* ... Line 19 ... */
  gotoxy(3,setup_item[4].y) ;
  cputs("Initialize Printer Before Configuring or Downloading Font ") ;
  post_yn() ;
  cputs("=>") ;
  /* ... Line 23 ... */
  gotoxy(3,23) ;
  other_color("ESC",white) ;
  cputs(" to exit  ") ;
  other_color("Alt-I",white) ;
  cputs("nitialize Printer  ") ;
  other_color("Alt-S",white) ;
  cputs("ave Data File  ") ;
  other_color("Alt-F",white) ;
  cputs("onts  ") ;
  other_color("Alt-M",white) ;
  cputs("ain") ;
}


/*============================================================================
FUNCTION NAME:  file2prn
PURPOSE:  Outputs a file to the printer
RETURNS:  int = 0, file printed okay
                1, error in printing file
ARGUMENTS:  char *fname = filename to print
GLOBAL VARIABLES:  init_code, initialize, stdprn
FUNCTIONS USED:  fclose, fgetc, filelength, fileno, flushall, fopen, fprintf,
                 message, putc, sleep
----------------------------------------------------------------------------*/
int file2prn ( char *fname )
{
char c ; /* the character just read */
FILE *infile ; /* file pointer to the input file */
int infileno ; /* file handle for infile */
int return_value ;  /* errorcode passed to calling routine */
long i ; /* a counter */
long infilelen ; /* file length of infile in bytes */

  return_value = 1 ;
  infile = fopen(fname,"rb") ;
  if ( infile == NULL ) /* error opening file */
  {
    message("Error Opening File") ;
    sleep(2) ;
    message("") ;
  }
  else
  {
    if ( prn_status(prn_port) == 0 ) /* printer ready */
    {
      flushall() ;
      message("Please Wait -- Downloading Font") ;
      infileno = fileno(infile) ;
      infilelen = filelength(infileno) ;
      if ( initialize == 'Y' )
        fprintf(stdprn,"%s",init_code) ; /* initialize printer */
      for ( i=0 ; i<infilelen ; i++ )
      {
        if ( bioskey(1) ) /* key pressed */
        {
          if ( bioskey(0) == 0x011B ) /* escape */
          {
            message("") ;
            message("Download Terminated") ;
            sleep(2) ;
            break ;
          }
        }
        else
        {
          c = fgetc(infile) ;
          putc(c,stdprn) ;
        }
      }
      return_value = 0 ;
      message("") ;
    }
  }
  fclose(infile) ;
  return(return_value) ;
}


/*============================================================================
FUNCTION NAME:  font_menu
PURPOSE:  Main control function for the Font Menu
RETURNS:  void
ARGUMENTS:  void
GLOBAL VARIABLES: choice, font_tbl, filename, font_dir, no_fonts, scan_codes
FUNCTIONS USED:  display_font_menu, file2prn, load_font_tbl, message,
                 position, sleep, strcat, strcpy, user_input
----------------------------------------------------------------------------*/
void font_menu ( void )
{
#define ITEMS_A_COL 16
struct BYTES { BYTE lo,hi ; } ;
struct UNSINT { unsigned int uns ; } ;
union WORD { struct BYTES  byte ;
             struct UNSINT word ; } key ;
int font_num, old_font_num ;

  scan_codes[1] = 0x8000 ; /* disable alt-i */
  scan_codes[4] = 0xA980 ; /* enable left & right arrows */
  display_font_menu() ;
  load_font_tbl() ;
  if ( no_fonts == 0 )
  {
    message("There are no font files") ;
    sleep(2) ;
    message("") ;
  }
  else /* display fonts */
  {
    for (font_num=0 ; font_num<no_fonts ; font_num++)
    {
      gotoxy(font_tbl[font_num].x,font_tbl[font_num].y) ;
      cputs(font_tbl[font_num].fontname) ;
    }
    font_num = 0 ;
    position(-1,font_num) ;
  }
  do
  {
    key.word.uns = user_input(NULL,NULL) ;
    old_font_num = font_num ;
    switch ( key.byte.hi )
    {
      case 0xFF : choice = ESC ; /* Escape */
                  break ;
      case 0x32 : choice = 'm' ; /* Alt-m, Main Menu */
                  break ;
      case 0x1F : choice = 's' ; /* Alt-s, Setup Menu */
                  break ;
      default   : if ( no_fonts > 0 )
                    switch ( key.byte.hi )
                    {
                      case 0x20 :   /* Alt-d, download font */
                                  strcpy(filename,font_dir) ;
                                  strcat(filename,font_tbl[font_num].fontname);
                                  strcat(filename,".") ;
                                  strcat(filename,font_ext) ;
                                  file2prn(filename) ;
                                  break ;
                      case 0x47 : font_num = 0 ; /* home */
                                  position(old_font_num,font_num) ;
                                  break ;
                      case 0x4F : font_num = no_fonts - 1 ; /* end */
                                  position(old_font_num,font_num) ;
                                  break ;
                      case 0x48 : font_num-- ; /* up arrow */
                                  if ( font_num < 0 )
                                    font_num = no_fonts - 1 ;
                                  position(old_font_num,font_num) ;
                                  break ;
                      case 0x50 :   /* down arrow */
                                  font_num = ++font_num * (font_num<no_fonts) ;
                                  position(old_font_num,font_num) ;
                                  break ;
                      case 0x4B :   /* left arrow */
                                  if ( (font_num-ITEMS_A_COL) >= 0 )
                                  {
                                    font_num -= ITEMS_A_COL ;
                                    position(old_font_num,font_num) ;
                                  }
                                  break ;
                      case 0x4D :   /* right arrow */
                                  if ( (font_num+ITEMS_A_COL) < no_fonts )
                                  {
                                    font_num += ITEMS_A_COL ;
                                    position(old_font_num,font_num) ;
                                  }
                                  break ;
                      default   : break ;
                    }
                  break ;
    }
  } while ( choice == 'f' ) ;
}


/*============================================================================
FUNCTION NAME:  horz_line
PURPOSE:  Puts a double horizontal line across the screen
RETURNS:  void
ARGUMENTS:  void
GLOBAL VARIABLES:  None
FUNCTIONS USED:  putch
----------------------------------------------------------------------------*/
void horz_line ( void )
{
int col ;

  for ( col=2 ; col<80 ; col++ )
    putch(HORZ) ;
}


/*============================================================================
FUNCTION NAME:  input_error
PURPOSE:  Posts an error message at bottom of main menu when user input
          is invalid
RETURNS:  int = 1, error message has been processed
ARGUMENTS:  void
GLOBAL VARIABLES:  None
FUNCTIONS USED:  message, sleep
----------------------------------------------------------------------------*/
int input_error ( void )
{
  message("Invalid Input") ;
  sleep(2) ;
  message("") ;
  return(1) ;
}


/*============================================================================
FUNCTION NAME:  load_font_tbl
PURPOSE:  Loads the array font_tbl with the filenames (minus the exten-
          sions) of up to 80 font files contained in the font directory
          and sorts the filemanes in ascending order.
RETURNS:  void
ARGUMENTS:  void
GLOBAL VARIABLES:  chg_dir, ffblk, filename, font_dir, font_tbl, no_fonts
FUNCTIONS USED:  findfirst, findnext, strcat, strcmp, strcpy,
----------------------------------------------------------------------------*/
void load_font_tbl ( void )
{
struct ffblk Ffblk ;
char temp[9] ;
int done, font_num, hi, i, lo, nent, ptr ;
int x = 6 ;
int y = 6 ;

  if ( chg_dir )
  {
    chg_dir = 0 ;
    no_fonts = 0 ;
    strcpy(filename,font_dir) ;
    strcat(filename,"*.") ;
    strcat(filename,font_ext) ;
    done = findfirst (filename,&Ffblk,0) ;
    while ( !done && no_fonts<=80 )
    {
      for ( i=0 ; i<8 ; i++ ) /* load font_tbl */
      {
        if ((Ffblk.ff_name[i]==' ') || (Ffblk.ff_name[i]=='.'))
          break ;
        font_tbl[no_fonts].fontname[i] = Ffblk.ff_name[i] ;
      }
      font_tbl[no_fonts].fontname[i] = NULL ; /* terminate with null */
      no_fonts++ ;
      done = findnext (&Ffblk) ; /* get name of next font file */
    }
    /* ... Sort font table in ascending order - double selection sort ... */
    i = 0 ;
    nent = no_fonts - 1 ;
    while ( i < nent )
    {
      lo = i ;
      hi = i ;
      ptr = i + 1 ;
      do
      {
        if ( strcmp(font_tbl[ptr].fontname,font_tbl[lo].fontname) < 0 )
          lo = ptr ; /* ptr < lo */
        else
        {
          if ( strcmp(font_tbl[ptr].fontname,font_tbl[hi].fontname) > 0 )
            hi = ptr ; /* ptr > hi */
        }
        ptr++ ;
      } while ( ptr <= nent ) ;
      if ( lo != i )
      {
        strcpy (temp,font_tbl[lo].fontname) ; /* Swap lo and i entries */
        strcpy (font_tbl[lo].fontname,font_tbl[i].fontname) ;
        strcpy (font_tbl[i].fontname,temp) ;
        if ( hi == i )
          hi = lo ;
      }
      if ( hi != nent )
      {
        strcpy (temp,font_tbl[hi].fontname) ; /* Swap hi and nent entries */
        strcpy (font_tbl[hi].fontname,font_tbl[nent].fontname) ;
        strcpy (font_tbl[nent].fontname,temp) ;
      }
      i++ ;
      nent-- ;
    }
    for (font_num=0 ; font_num<no_fonts ; font_num++)
    {
      if ( y > 21 )
      {
        x += 14 ;
        y = 6 ;
      }
      font_tbl[font_num].x = x ;
      font_tbl[font_num].y = y ;
      font_tbl[font_num].len = strlen(font_tbl[font_num].fontname) ;
      ++y ;
    }
  }
}


/*============================================================================
FUNCTION NAME:  main_menu
PURPOSE:  The control routine for the main menu.  Displays the main menu
          and handles updates and configuration of the printer.
RETURNS:  void
ARGUMENTS:  void
GLOBAL VARIABLES:  choice, defcodes, init_code, initialize, main_item,
                   max_margin, prn_port, scan_codes, stdprn, workrec
FUNCTIONS USED:  assign_default, display_main_menu, fprintf, gotoxy,
                 input_error, message, prn_status, sscanf, sleep, stricmp,
                 update_main
----------------------------------------------------------------------------*/
void main_menu ( void )
{
BYTE *workrec_ptr ;
char instr[4] ;
char *no = "N" ;
char *yes = "Y" ;
int error, i, item_num, num ;
unsigned int return_value ;

  max_margin = ( carriage==10 ? 160:255 ) ;
  scan_codes[1] = 0x8000 ; /* disable alt-i */
  scan_codes[4] = 0x8180 ; /* disable left & right arrows */
  workrec_ptr = &workrec.prop_char[0] ;
  display_main_menu() ;
  for (item_num=0 ; item_num<NO_MAIN_ITEMS ; item_num++)
    update_main(item_num) ;
  item_num = 0 ;
  do
  {
    gotoxy(main_item[item_num].x,main_item[item_num].y) ;
    return_value = user_input(instr,main_item[item_num].max_bytes) ;
    if ( return_value == 0xFFFF ) /* escape pressed */
      choice = ESC ;
    else
    {
      if ( return_value!=0 && return_value<0x0100 ) /* ascii inputed */
      {
        error = 0 ;
        switch ( item_num )
        {
          case  0 : sscanf(instr,"%d",&num) ; /* Characters Per Inch */
                    if ( num==10 || num==12 || num==15 )
                    {
                      switch ( num )
                      {
                        case 10 : workrec.pitch[1] = 'P' ; /* pica */
                                  break ;
                        case 12 : workrec.pitch[1] = 'M' ; /* elite */
                                  break ;
                        case 15 : workrec.pitch[1] = 'g' ; /* 15 CPI */
                                  break ;
                        default : break ;
                      }
                      workrec.prop_char[2] = 0 ; /* no prop char */
                      update_main(4) ;
                    }
                    else
                      error = input_error() ;
                    break ;
          case  1 : if ( stricmp(instr,yes) == 0 ) /* Double Width */
                      workrec.dw[2] = 1 ; /* select */
                    else
                    {
                      if ( stricmp(instr,no) == 0 )
                        workrec.dw[2] = 0 ; /* cancel */
                      else
                        error = input_error() ;
                    }
                    break ;
          case  2 : if ( stricmp(instr,yes) == 0 ) /* Condensed */
                      workrec.cond[0] = 15 ; /* select */
                    else
                    {
                      if ( stricmp(instr,no) == 0 )
                        workrec.cond[0] = 18 ; /* cancel */
                      else
                        error = input_error() ;
                    }
                    break ;
          case  3 : if ( stricmp(instr,yes) == 0 ) /* Italics */
                      workrec.italic[1] = '4' ; /* select */
                    else
                    {
                      if ( stricmp(instr,no) == 0 )
                        workrec.italic[1] = '5' ; /* cancel */
                      else
                        error = input_error() ;
                    }
                    break ;
          case  4 : if ( stricmp(instr,yes) == 0 ) /* Proportional Char Set */
                      workrec.prop_char[2] = 1 ; /* select */
                    else
                    {
                      if ( stricmp(instr,no) == 0 )
                        workrec.prop_char[2] = 0 ; /* cancel */
                      else
                      {
                        error = input_error() ;
                        break ;
                      }
                    }
                    update_main(0) ;
                    break ;
          case  5 : if ( stricmp(instr,"SUP") == 0 ) /* Super / Subscript */
                    {
                      workrec.sup_sub[1] = 'S' ; /* select superscript */
                      workrec.sup_sub[2] = 0 ;
                    }
                    else
                    {
                      if ( stricmp(instr,"SUB") == 0 )
                      {
                        workrec.sup_sub[1] = 'S' ; /* select subscript */
                        workrec.sup_sub[2] = 1 ;
                      }
                      else
                      {
                        if ( stricmp(instr,no) == 0 )
                        {
                          workrec.sup_sub[1] = 'T' ; /* cancel both */
                          workrec.sup_sub[2] = NULL ;
                        }
                        else
                          error = input_error() ;
                      }
                    }
                    break ;
          case  6 : sscanf (instr,"%d",&num) ; /* Left Margin */
                    if ( num>=0 && num<=max_margin )
                      workrec.left_mar[2] = num ;
                    else
                      error = input_error() ;
                    break ;
          case  7 : sscanf (instr,"%d",&num) ; /* Right Margin */
                   if ( num>=0 && num<=max_margin && num>workrec.left_mar[2] )
                      workrec.right_mar[2] = num ;
                    else
                      error = input_error() ;
                    break ;
          case  8 : if ( stricmp(instr,"D") == 0 ) /* Draft/Letter Quality */
                      workrec.draft_lq[2] = 0 ; /* select draft quality */
                    else
                    {
                      if ( stricmp(instr,"LQ") == 0 )
                        workrec.draft_lq[2] = 1 ; /* select letter quality */
                      else
                        error = input_error() ;
                    }
                    break ;
          case  9 : if ( stricmp(instr,yes) == 0 ) /* Double Strike */
                      workrec.ds[1] = 'G' ; /* select */
                    else
                    {
                      if ( stricmp(instr,no) == 0 )
                        workrec.ds[1] = 'H' ; /* cancel */
                      else
                        error = input_error() ;
                    }
                    break ;
          case 10 : if ( stricmp(instr,yes) == 0 ) /* Emphasized */
                      workrec.emph[1] = 'E' ; /* select */
                    else
                    {
                      if ( stricmp(instr,no) == 0 )
                        workrec.emph[1] = 'F' ; /* cancel */
                      else
                        error = input_error() ;
                    }
                    break ;
          case 11 : if ( stricmp(instr,yes) == 0 ) /* Unidirectional */
                      workrec.uni[2] = 1 ; /* select */
                    else
                    {
                      if ( stricmp(instr,no) == 0 )
                        workrec.uni[2] = 0 ; /* cancel */
                      else
                        error = input_error() ;
                    }
                    break ;
          case 12 : if ( stricmp(instr,yes) == 0 ) /* Half-Speed */
                      workrec.half_speed[2] = 1 ; /* select */
                    else
                    {
                      if ( stricmp(instr,no) == 0 )
                        workrec.half_speed[2] = 0 ; /* cancel */
                      else
                        error = input_error() ;
                    }
                    break ;
          case 13 : sscanf (instr,"%d",&num) ; /* Page Length */
                    if ( num>=1 && num<=32 )
                      workrec.length[3] = num ;
                    else
                      error = input_error() ;
                    break ;
          case 14 : sscanf (instr,"%d",&num) ; /* Lines per Page */
                    if ( num>=1 && num<=127 )
                      workrec.lines[2] = num ;
                    else
                      error = input_error() ;
                    break ;
          case 15 : if ( stricmp(instr,yes) == 0 ) /* 1/8" Line Spacing */
                    {
                      workrec.line_spacing[1] = '0' ; /* set to 1/8" */
                      workrec.line_spacing[2] = NULL ;
                      update_main (16) ; /* ensure line spacing is 'NA' */
                    }
                    else
                    {
                      if ( stricmp(instr,no) == 0 )
                      {
                        if ( workrec.line_spacing[1] != 'A' )
                        {
                          assign_default (16) ; /* set to default value */
                          update_main (16) ;
                        }
                      }
                      else
                        error = input_error() ;
                    }
                    break ;
          case 16 : sscanf (instr,"%d",&num) ; /* Line Spacing n/60" */
                    if ( num>=1 && num<=85 )
                    {
                      workrec.line_spacing[1] = 'A' ;
                      workrec.line_spacing[2] = num ;
                      update_main (15) ; /* ensure 1/8" line spacing is 'N' */
                    }
                    else
                      error = input_error() ;
                    break ;
          case 17 : sscanf (instr,"%d",&num) ; /* Skip Lines Before Perf */
                    if ( num>=0 && num<=255 )
                      workrec.skip_perf[2] = num ;
                    else
                      error = input_error() ;
                    break ;
          default : break ;
        }
        update_main(item_num) ;
        if ( !error )
          item_num = ++item_num * (item_num<NO_MAIN_ITEMS) ;
      }
      else /* non-ascii input */
      {
        update_main(item_num) ;
        switch ( return_value )
        {
          case 0x0000 :   /* return pressed on first character */
          case 0x5000 :   /* down arrow */
                        item_num = ++item_num * (item_num<NO_MAIN_ITEMS) ;
                        break ;
          case 0x2E00 : if ( prn_status(prn_port) == 0 ) /* Alt-c, conf prntr */
                        {
                          message ("Please Wait -- Configuring Printer") ;
                          if ( initialize == 'Y' )
                            fprintf(stdprn,"%s",init_code) ; /* init prn */
                          for (i=0 ; i<46 ; i++) /* send workrec to printer */
                            putc (*(workrec_ptr+i),stdprn) ;
                          sleep(1) ;
                          message ("") ;
                        }
                        break ;
          case 0x2000 : for (i=0 ; i<NO_MAIN_ITEMS ; i++) /* Alt-d, defaults */
                        {
                          assign_default(i) ;
                          update_main(i) ;
                        }
                        item_num = 0 ;
                        break ;
          case 0x2100 : choice = 'f' ; /* Alt-f, font menu */
                        break ;
          case 0x1F00 : choice = 's' ; /* Alt-s, setup menu */
                        break ;
          case 0x4800 : item_num-- ; /* up arrow */
                        if ( item_num < 0 )
                          item_num = NO_MAIN_ITEMS - 1 ;
                        break ;
          case 0x4700 : item_num = 0 ; /* home */
                        break ;
          case 0x4F00 : item_num = NO_MAIN_ITEMS - 1 ; /* end */
                        break ;
          default     : break ;
        }
      }
    }
  } while ( choice == 'm' ) ;
}


/*============================================================================
FUNCTION NAME:  menu_frame
PURPOSE:  Clears the screen and displays the menu frame with title.
          This frame is used by all three menus.
RETURNS:  void
ARGUMENTS:  char *title = title to be displayed in the menu frame
GLOBAL VARIABLES:  None
FUNCTIONS USED:  clrscr, gotoxy, horz_line, other_color, outside_edges,
                 putch, tee_line
----------------------------------------------------------------------------*/
void menu_frame ( char *title )
{
int line, title_col ;

  clrscr() ;
  /* ... Line 1 ... */
  putch(UL) ; /* Upper left-hand corner */
  horz_line() ; /* Double horizontal line */
  putch(UR) ; /* Upper right-hand corner */
  /* ... Lines 2 thru 4 ... */
  for ( line=2 ; line<5 ; line++ ) /* Outside edges */
    outside_edges(line) ;
  /* ... Line 5 ... */
  tee_line(5) ; /* Tee'd line */
  /* ... Lines 6 thru 21 ... */
  for ( line=6 ; line<22 ; line++ ) /* Outside edges */
    outside_edges(line) ;
  /* ... Line 22 ... */
  tee_line(22) ; /* Tee'd line */
  /* ... Line 23 ... */
  outside_edges(23) ;
  /* ... Line 24 ... */
  putch(LL) ; /* Lower left-hand corner */
  horz_line() ; /* Double horizontal line */
  putch(LR) ; /* Lower right-hand corner */
  /* ... Display Title ... */
  title_col = (80 - strlen(title))/2 + 1;
  gotoxy(title_col,3) ;
  other_color(title,white) ; /* Display title in white */
}


/*============================================================================
FUNCTION NAME:  message
PURPOSE:  Displays a message centered on line 25 and beeps.  If *msg
          is a null string, then line 25 message area is cleared.
RETURNS:  void
ARGUMENTS:  char *msg = message to display on line 25
GLOBAL VARIABLES:  hi_red, text_info, white
FUNCTIONS USED:  clreol, cputs, gettextinfo, gotoxy, putch, strlen, textattr,
                 textbackground, textcolor
----------------------------------------------------------------------------*/
void message ( char *msg )
{
static struct text_info Text_info ;
int msg_col ;

  if ( strlen(msg) > 0 )
  {
    gettextinfo(&Text_info) ; /* Get current screen attribute */
    msg_col = (80 - strlen(msg))/2 ;
    gotoxy(msg_col,25) ;
    textbackground(hi_red) ;
    textcolor(white) ;
    putch(' ') ;
    cputs(msg) ;
    cputs(" ") ;
  }
  else
  {
    textattr(Text_info.attribute) ; /* Change attribute to previous colors */
    gotoxy(1,25) ;
    clreol() ;
  }
}


/*============================================================================
FUNCTION NAME:  no_yes
PURPOSE:  Posts 'N' or 'Y' for main menu item
RETURNS:  void
ARGUMENTS:  char *workrec_ptr = pointer to workrec character to compare
            char no_value     = character equivalent to 'N'
GLOBAL VARIABLES:  None
FUNCTIONS USED:  putch
----------------------------------------------------------------------------*/
void no_yes ( BYTE *workrec_ptr, char no_value )
{
  if ( *workrec_ptr == no_value )
    putch('N') ;
  else
    putch('Y') ;
}


/*============================================================================
FUNCTION NAME:  other_color
PURPOSE:  Displays the passed string in the color specified and then
          restores the previous text color
RETURNS:  void
ARGUMENTS:  char *str = pointer of string to display
            int color = color to display string
GLOBAL VARIABLES:  None
FUNCTIONS USED:  cputs, gettextinfo, textcolor
----------------------------------------------------------------------------*/
void other_color ( char *str, int color )
{
struct text_info Text_info ;

  gettextinfo(&Text_info) ; /* Get current screen attribute */
  textcolor(color) ; /* Color of string */
  cputs(str) ;
  textcolor(Text_info.attribute & 0x0F) ; /* Change text to previous color */
}


/*============================================================================
FUNCTION NAME:  outside_edges
PURPOSE:  Puts a double vertical line in columns 1 and 80 on the
          specified line
RETURNS:  void
ARGUMENTS:  int line = line on which to put the vertical lines
GLOBAL VARIABLES:  None
FUNCTIONS USED:  gotoxy, putch
----------------------------------------------------------------------------*/
void outside_edges ( int line )
{
  gotoxy(1,line) ;
  putch(VERT) ;
  gotoxy(80,line) ;
  putch(VERT) ;
}


/*============================================================================
FUNCTION NAME:  position
PURPOSE:  Goes to the column and row specified in font_tbl and un-
          highlights the old entry and highlights the new entry.  If old
          is less then 0, then there is no old entry to un-highlight.
RETURNS:  void
ARGUMENTS:  int old = previous font_tbl entry to un-highlight
            int new = new font_tbl entry to highlight
GLOBAL VARIABLES:  font_tbl
FUNCTIONS USED:  REVID
----------------------------------------------------------------------------*/
void position ( int old, int new )
{
  if ( old >= 0 )
    REVID(font_tbl[old].x,font_tbl[old].y,font_tbl[old].len) ;
  REVID(font_tbl[new].x,font_tbl[new].y,font_tbl[new].len) ;
}


/*============================================================================
FUNCTION NAME: post_limits
PURPOSE:  Posts (n-nnn) in main menu
RETURNS:  void
ARGUMENTS:  char *first  = first number to post
            char *second = second number to post
GLOBAL VARIABLES:  white
FUNCTIONS USED:  cputs, other_color, putch
----------------------------------------------------------------------------*/
void post_limits ( char *first, char *second )
{
  putch('(') ;
  other_color(first,white) ; /* high-intensity white */
  putch('-') ;
  other_color(second,white) ;
  cputs(") ") ;
}


/*============================================================================
FUNCTION NAME:  post_yn
PURPOSE:  Posts (Y/N) in main menu
RETURNS:  void
ARGUMENTS:  void
GLOBAL VARIABLES:  white
FUNCTIONS USED:  cputs, other_color, putch
----------------------------------------------------------------------------*/
void post_yn ( void )
{
  putch('(') ;
  other_color("Y",white) ; /* high-intensity white */
  putch('/') ;
  other_color("N",white) ;
  cputs(") ") ;
}


/*============================================================================
FUNCTION NAME:  prn_status
PURPOSE:  Checks status of printer, posts message if not ready
RETURNS:  int = 0, printer is ready
                1, printer is not ready
ARGUMENTS:  int port = 0, printer port LPT1
                       1, printer port LPT2
GLOBAL VARIABLES:  None
FUNCTIONS USED:  biosprint, message, sleep
----------------------------------------------------------------------------*/
int prn_status ( int port )
{
int return_value = 0 ;

  if ( biosprint(2,NULL,port) != 0x0090 ) /* printer not ready */
  {
    message("Error Accessing Printer -- Check Printer") ;
    sleep(2) ;
    message("") ;
    return_value++ ;
  }
  return(return_value) ;
}


/*============================================================================
FUNCTION NAME:  setup_menu
PURPOSE:  The control routine for the setup menu.  Allows user specified
          directory for font files and options to save a configuration
          file.
RETURNS:  void
ARGUMENTS:  void
GLOBAL VARIABLES:  carriage, chg_dir, choice, font_dir, init_code, initialize,
                   main_item, prn_port, scan_codes, setup_item, stdprn, workrec
FUNCTIONS USED:  config, display_setup_menu, fflush, fprintf, gotoxy,
                 input_error, message, prn_status, sleep, sscanf, strcpy,
                 stricmp, strlen, update_setup, user_input
----------------------------------------------------------------------------*/
void setup_menu ( void )
{
struct BYTES { BYTE lo,hi ; } ;
struct UNSINT { unsigned int uns ; } ;
union WORD { struct BYTES  byte ;
             struct UNSINT word ; } key ;
char instr[53] ;
int error, instr_len, item_num, num ;

  scan_codes[1] = 0x8080 ; /* enable alt-i */
  scan_codes[4] = 0x8180 ; /* disable left & right arrows */
  display_setup_menu() ;
  for (item_num=0 ; item_num<NO_SETUP_ITEMS ; item_num++)
    update_setup(item_num) ;
  item_num = 0 ;
  do
  {
    gotoxy(setup_item[item_num].x,setup_item[item_num].y) ;
    key.word.uns = user_input(instr,setup_item[item_num].max_bytes) ;
    if ( key.word.uns == 0xFFFF ) /* escape pressed */
      choice = ESC ;
    else
    {
      if ( key.byte.lo !=0 ) /* ascii inputed */
      {
        error = 0 ;
        switch ( item_num )
        {
          case  0 : instr_len = strlen(instr) ;/* font files directory */
                    if ( instr[instr_len-1] != BACKSLASH )
                    {
                      instr[instr_len] = BACKSLASH ;
                      instr[instr_len+1] = NULL ;
                    }
                    strcpy(font_dir,instr) ;
                    chg_dir = 1 ;
                    break ;
          case 1 :  strcpy(font_ext,instr) ; /* font files extension */
                    chg_dir = 1 ;
                    break ;
          case 2 :  sscanf(instr,"%d",&num) ; /* printer port */
                    if ( num==1 || num==2 )
                      prn_port = num - 1 ;
                    else
                      error = input_error() ;
                    break ;
          case 3 :  sscanf(instr,"%d",&num) ; /* carriage size */
                    switch ( num )
                    {
                      case 10 : carriage = num ;
                                if ( workrec.left_mar[2] > 160 )
                                  workrec.left_mar[2] = 0 ;
                                if ( workrec.right_mar[2] > 160 )
                                  workrec.right_mar[2] = 80 ;
                                main_item[7].code[2] = 80 ;
                                break ;
                      case 15 : carriage = num ;
                                main_item[7].code[2] = 136 ;
                                break ;
                      default : error = input_error() ;
                                break ;
                     }
                     break ;
          case 4 :  if ( stricmp(instr,"Y") == 0 ) /* init prn option */
                      initialize = 'Y' ;
                    else
                    {
                      if ( stricmp(instr,"N") == 0 )
                        initialize = 'N' ;
                      else
                        error = input_error() ;
                    }
                    break ;
          default : break ;
        }
        update_setup(item_num) ;
        if ( !error )
          item_num = ++item_num * (item_num<NO_SETUP_ITEMS) ;
      }
      else /* non-ascii input */
      {
        update_setup(item_num) ;
        switch ( key.byte.hi )
        {
          case 0xFF : choice = ESC ; /* Escape */
                      break ;
          case 0x21 : choice = 'f' ; /* Alt-f, Fonts Menu */
                      break ;
          case 0x32 : choice = 'm' ; /* Alt-m, Main Menu */
                      break ;
          case 0x17 : if ( prn_status(prn_port) == 0 ) /* Alt-i */
                      {
                        message("Please Wait -- Initializing Printer") ;
                        fprintf(stdprn,"%s",init_code) ; /* initialize prn */
                        fflush(stdprn) ;
                        sleep(1) ;
                        message("") ;
                      }
                      break ;
          case 0x1F : config('S') ; /* Alt-s, save data file */
                      break ;
          case 0x47 : item_num = 0 ; /* home */
                      break ;
          case 0x4F : item_num = NO_SETUP_ITEMS - 1 ; /* end */
                      break ;
          case 0x48 : item_num-- ; /* up arrow */
                      if ( item_num < 0 )
                        item_num = NO_SETUP_ITEMS - 1 ;
                      break ;
          case 0x00 :   /* just carriage return */
          case 0x50 :   /* down arrow */
                      item_num = ++item_num * (item_num<NO_SETUP_ITEMS) ;
                      break ;
          default   : break ;
        }
      }
    }
  } while ( choice == 's' ) ;
}


/*============================================================================
FUNCTION NAME:  tee_line
PURPOSE:  Puts a 'teed' line on the screen at the line specified
RETURNS:  void
ARGUMENTS:  int line = line number to put the teed line
GLOBAL VARIABLES:  None
FUNCTIONS USED:  gotoxy, horz_line, putch
----------------------------------------------------------------------------*/
void tee_line ( int line )
{
  gotoxy(1,line) ;
  putch(T2R) ; /* Tee-to-right character */
  horz_line() ; /* Double horizontal line */
  putch(T2L) ; /* Tee-to-left character */
}


/*============================================================================
FUNCTION NAME:  update_main
PURPOSE:  Reads the value in workrec and translates it for display in
          the main menu
RETURNS:  void
ARGUMENTS:  int item_num = item number to be translated and displayed.
GLOBAL VARIABLES:  main_item, workrec
FUNCTIONS USED:  cputs, gotoxy, highvideo, lowvideo, no_yes, sprintf
----------------------------------------------------------------------------*/
void update_main ( int item_num )
{
char outstr[4] ;

  highvideo() ;
  gotoxy(main_item[item_num].x,main_item[item_num].y) ;
  switch ( item_num )
  {
    case  0 : if ( workrec.prop_char[2] == 0 ) /* CPI */
                switch ( workrec.pitch[1] )
                {
                  case 'P' : cputs("10") ;
                             break ;
                  case 'M' : cputs("12") ;
                             break ;
                  case 'g' : cputs("15") ;
                             break ;
                  default  : break ;
                }
                else
                  cputs("NA") ;
                break ;
    case  1 : no_yes (&workrec.dw[2],0) ; /* Double Width */
              break ;
    case  2 : no_yes (&workrec.cond[0],18) ; /* Condensed */
              break ;
    case  3 : no_yes (&workrec.italic[1],'5') ; /* Italics */
              break ;
    case  4 : no_yes (&workrec.prop_char[2],0) ; /* Proportional Char Set */
              break ;
    case  5 : if ( workrec.sup_sub[1] == 'T' ) /* Superscript/Subscript */
                cputs("N  ") ;
              else
              {
                if ( workrec.sup_sub[2] == 0 )
                  cputs("SUP") ;
                else
                  cputs("SUB") ;
              }
              break ;
    case  6 : sprintf(outstr,"%-3d",workrec.left_mar[2]) ; /* Left Margin */
              cputs(outstr) ;
              break ;
    case  7 : sprintf(outstr,"%-3d",workrec.right_mar[2]); /* Right Margin */
              cputs(outstr) ;
              break ;
    case  8 : if ( workrec.draft_lq[2] == 0 ) /* Draft/Letter Quality */
                cputs("D ") ;
              else
                cputs("LQ") ;
              break ;
    case  9 : no_yes(&workrec.ds[1],'H') ; /* Double Strike */
              break ;
    case 10 : no_yes(&workrec.emph[1],'F') ; /* Emphasized */
              break ;
    case 11 : no_yes(&workrec.uni[2],0) ; /* Unidirectional Print */
              break ;
    case 12 : no_yes(&workrec.half_speed[2],0) ; /* Half-Speed */
              break ;
    case 13 : sprintf(outstr,"%-2d",workrec.length[3]) ; /* Page Length */
              cputs(outstr) ;
              break ;
    case 14 : sprintf(outstr,"%-3d",workrec.lines[2]); /* Lines per Page */
              cputs(outstr) ;
              break ;
    case 15 : no_yes(&workrec.line_spacing[1],'A') ; /* 1/8" Line Spacing */
              break ;
    case 16 : if ( workrec.line_spacing[1] == 'A' ) /* Line Spacing n/60" */
              {
                sprintf(outstr,"%-2d",workrec.line_spacing[2]) ;
                cputs(outstr) ;
              }
              else
                cputs("NA") ;
              break ;
    case 17 : sprintf(outstr,"%-3d",workrec.skip_perf[2]); /* Skip Perf */
              cputs(outstr) ;
              break ;
    default : break ;
  }
  lowvideo() ;
}


/*============================================================================
FUNCTION NAME:  update_setup
PURPOSE:  Reads the variable pertaining to the setup menu item number and
          translates it for display in the setup menu
RETURNS:  void
ARGUMENTS:  int item_num = item number to update
GLOBAL VARIABLES:  carriage, font_dir, font_ext, initialize, prn_port,
                   setup_item
FUNCTIONS USED:  cputs, gotoxy, highvideo, lowvideo, putch, sprintf, strupr
----------------------------------------------------------------------------*/
void update_setup ( int item_num )
{
char outstr[3] ;
int i ;

  highvideo() ;
  gotoxy(setup_item[item_num].x,setup_item[item_num].y) ;
  switch ( item_num )
  {
    case  0 : strupr(font_dir) ; /* font files directory */
              for (i=0 ; i<setup_item[item_num].max_bytes ; i++ )
                putch(' ') ;
              gotoxy(setup_item[item_num].x,setup_item[item_num].y) ;
              cputs(font_dir) ;
              break ;
    case  1 : strupr(font_ext) ; /* font files extension */
              for (i=0 ; i<setup_item[item_num].max_bytes ; i++ )
                putch(' ') ;
              gotoxy(setup_item[item_num].x,setup_item[item_num].y) ;
              cputs(font_ext) ;
              break ;
    case  2 : sprintf(outstr,"%d",prn_port+1) ; /* printer port */
              cputs(outstr) ;
              break ;
    case  3 : sprintf(outstr,"%2d",carriage) ; /* carriage size */
              cputs(outstr) ;
              break ;
    case  4 : putch(initialize) ; /* initialize printer option */
              break ;
    default : break ;
  }
  lowvideo() ;
}


/*============================================================================
FUNCTION NAME:  unsigned int user_input (char *str, int max_bytes)
PURPOSE:  Keyboard trap routine for main menu & setup menu
RETURNS:  unsigned int =  < 0x0100 = number of bytes entered
                         >= 0x0100 = allowable key other than ascii was pressed
                          = 0xFFFF = escape pressed
ARGUMENTS:  char *str     = input string from calling routine
            int max_bytes = maximum number of bytes to input
                            if 0, don't process ascii characters
GLOBAL VARIABLES:  scan_codes
FUNCTIONS USED:  bioskey, cursor, putch, strcpy
----------------------------------------------------------------------------*/
unsigned int user_input ( char *str, int max_bytes )
{
struct BYTES { BYTE lo,hi ; } ;
struct UNSINT { unsigned int uns ; } ;
union WORD { struct BYTES  byte ;
             struct UNSINT word ; } key ;
char input[81] ;
unsigned int bit, return_value ;
int done = 0 ;
int ndx = 0 ;
int i ;

  max_bytes = min(max_bytes,80) ;
  if ( max_bytes > 0 )
    cursor(ON,0x0707) ;
  while ( !done )
  {
    while ( bioskey(1) == 0 ) { } /* wait until key is pressed */
    key.word.uns = bioskey(0) ; /* get key value and clear keyboard buffer */
    if ( !key.byte.lo ) /* not an ascii character */
    {
      bit = 0x0001 << (key.byte.hi%0x10) ;
      if ( scan_codes[key.byte.hi/0x10] & bit ) /* if bit set - process */
      {
        return_value = key.word.uns & 0xFF00 ;
        done++ ;
      }
    }
    else
    {
      if ( key.byte.lo>='!'&& key.byte.lo<='~' && ndx<max_bytes )
      {                                        /* allowable ascii characters */
        input[ndx] = key.byte.lo ;
        putch(key.byte.lo) ;
        ndx++ ;
        if ( ndx == 1 )
        {
          for (i=1 ; i<max_bytes ; i++)
            putch(' ') ;
          for ( ; i>1 ; i--)
            putch(8) ;
        }
      }
      else /* check for other permissible keystrokes */
      {
        switch ( key.byte.lo )
        {
          case 0x08 : if ( ndx>0 && max_bytes>0 ) /* backspace */
                      {
                        ndx-- ;
                        putch(8) ;
                        putch(' ') ;
                        putch(8) ;
                      }
                      break ;
          case 0x0D : if ( max_bytes > 0 ) /* return */
                      {
                        input[ndx] = NULL ;
                        if ( ndx > 0 )
                          strcpy(str,input) ;
                        return_value = ndx ;
                        done++ ;
                      }
                      break ;
          case 0x1B : return_value = ( ndx==0 ? 0xFFFF:0x0000 ) ; /* escape */
                      done++ ;
                      break ;
          default   : break ; /* ignore remaining characters */
        }
      }
    }
  }
  if ( max_bytes > 0 )
    cursor(OFF,NULL) ;
  return(return_value) ;
}
